from cone import sql
from cone.app import RemoteAddrFilter
from cone.sql import testing
from node.tests import NodeTestCase
from pyramid.paster import get_app
from sqlalchemy.orm.session import Session
import os
import shutil
import tempfile


def temp_directory(fn):
    """Decorator for tests needing a temporary directory.
    """
    def wrapper(*a, **kw):
        tempdir = tempfile.mkdtemp()
        kw['tempdir'] = tempdir
        try:
            fn(*a, **kw)
        finally:
            shutil.rmtree(tempdir)
    return wrapper


app_config = """
[app:my_app]
use = egg:cone.app#main

tm.commit_veto = pyramid_tm.default_commit_veto

cone.plugins =
    cone.sql

sql.db.url = sqlite:///:memory:

[filter:remote_addr]
# for use behind nginx
use = egg:cone.app#remote_addr

[filter:session]
use = egg:cone.sql#session

[pipeline:main]
pipeline =
    remote_addr
    session
    my_app
"""


class TestSQL(NodeTestCase):
    layer = testing.sql_layer

    @temp_directory
    def test_wsgi(self, tempdir):
        # Write test application config
        config_path = os.path.join(tempdir, 'sql.ini')
        with open(config_path, 'w') as f:
            f.write(app_config)

        # Create WSGI app
        wsgi_app = get_app(config_path, 'main')
        self.assertTrue(isinstance(wsgi_app, RemoteAddrFilter))

        # Dummy WSGI environment
        environ = {
            'REQUEST_METHOD': 'GET',
            'PATH_INFO': '/'
        }

        # Dummy ``start_response`` callback
        class StartResponse(object):
            args = None

            def __call__(self, *args):
                self.args = args

        # Call WSGI app
        start_response = StartResponse()
        wsgi_app(environ, start_response)
        result = start_response.args
        self.assertEqual(result[0], '200 OK')
        self.assertEqual(
            result[1][0],
            ('Content-Type', 'text/html; charset=UTF-8')
        )
        self.assertEqual(result[1][1][0], 'Content-Length')

        # SQL session has been hooked up to environment
        self.assertTrue(isinstance(environ[sql.session_key], Session))



from cone import sql
from cone.app.testing import Security
from cone.app.ugm import ugm_backend
from cone.sql import get_session
from cone.sql import initialize_sql
from cone.sql import setup_session
from cone.sql import sql_session_setup
from sqlalchemy import create_engine
from sqlalchemy import event
from sqlalchemy.orm import sessionmaker
import os
import shutil
import tempfile


try:
    from cone.ugm import testing
    CONE_UGM_INSTALLED = True
except ImportError:
    CONE_UGM_INSTALLED = False


###############################################################################
# Test session setup handler and event listener
###############################################################################

# override to test if event listener gets called properly
test_after_flush = None


def after_flush(session, flush_context):
    """Test event listener.
    """
    if test_after_flush is not None:
        test_after_flush(session, flush_context)


@sql_session_setup
def bind_session_listener(session):
    """Test SQL session setup callback.
    """
    event.listen(session, 'after_flush', after_flush)


###############################################################################
# Test decorators
###############################################################################

class delete_table_records(object):

    def __init__(self, record_cls):
        self.record_cls = record_cls

    def __call__(self, fn):
        def wrapper(inst):
            try:
                fn(inst)
            finally:
                request = inst.layer.new_request()
                session = get_session(request)
                session.query(self.record_cls).delete()
                session.commit()
        return wrapper


###############################################################################
# Test layer
###############################################################################

class SQLLayer(Security):

    def make_app(self):
        kw = dict()
        plugins = ['cone.sql']
        if CONE_UGM_INSTALLED:
            plugins.append('cone.ugm')
            kw['ugm.backend'] = 'sql'
            kw['ugm.config'] = testing.ugm_config
            kw['ugm.localmanager_config'] = testing.localmanager_config
            kw['sql.binary_attrs'] = 'portrait'
        kw['cone.plugins'] = '\n'.join(plugins)
        super(SQLLayer, self).make_app(**kw)
        if CONE_UGM_INSTALLED:
            ugm_backend.initialize()

    def setUp(self, args=None):
        self.tempdir = tempfile.mkdtemp()
        super(SQLLayer, self).setUp()
        self.init_sql()
        self.new_request()

    def tearDown(self):
        super(SQLLayer, self).tearDown()
        shutil.rmtree(self.tempdir)

    def new_request(self, type=None, xhr=False):
        request = super(SQLLayer, self).new_request(type=type, xhr=xhr)
        request.environ['cone.sql.session'] = self.sql_session
        return request

    def init_sql(self):
        sql_backend = os.environ.get('CONE_SQL_TEST_BACKEND')
        # sqlite memory is default test backend
        if not sql_backend:  # pragma no cover
            engine = create_engine('sqlite:///:memory:', echo=False)
        # sqlite persistent in package folder for post mortem analysis
        elif sql_backend == 'sqlite':  # pragma no cover
            curdir = os.path.dirname(__file__)
            fname = "%s/test.db" % curdir
            if os.path.exists(fname):
                os.remove(fname)
            uri = "sqlite:///" + fname
            engine = create_engine(uri)
        # alternatively use postgresql - ditches db before start
        elif sql_backend == 'postgres':  # pragma no cover
            os.system("dropdb ugm; createdb ugm")
            engine = create_engine("postgresql:///ugm", echo=False)
        initialize_sql(engine)
        maker = sessionmaker(bind=engine)
        if sql.session_factory:  # pragma no cover
            sql.session_factory.maker = maker
        session = maker()
        setup_session(session)
        self.sql_session = session


sql_layer = SQLLayer()



from cone.sql import testing
from cone.ugm.tests import test_browser_actions
from cone.ugm.tests import test_browser_autoincrement
from cone.ugm.tests import test_browser_expires
from cone.ugm.tests import test_browser_group
from cone.ugm.tests import test_browser_groups
from cone.ugm.tests import test_browser_portrait
from cone.ugm.tests import test_browser_principal
from cone.ugm.tests import test_browser_remote
from cone.ugm.tests import test_browser_roles
from cone.ugm.tests import test_browser_root
from cone.ugm.tests import test_browser_settings
from cone.ugm.tests import test_browser_user
from cone.ugm.tests import test_browser_users
from cone.ugm.tests import test_browser_utils
from cone.ugm.tests import test_layout
from cone.ugm.tests import test_localmanager
from cone.ugm.tests import test_model_group
from cone.ugm.tests import test_model_groups
from cone.ugm.tests import test_model_user
from cone.ugm.tests import test_model_users
from cone.ugm.tests import test_settings
from cone.ugm.tests import test_utils


class TestLayout(test_layout.TestLayout):
    pass


class TestModelLocalmanager(test_localmanager.TestModelLocalmanager):
    layer = testing.sql_layer


class TestModelSettings(test_settings.TestModelSettings):
    layer = testing.sql_layer


class TestModelUtils(test_utils.TestModelUtils):
    layer = testing.sql_layer


class TestModelGroup(test_model_group.TestModelGroup):
    layer = testing.sql_layer


class TestModelGroups(test_model_groups.TestModelGroups):
    layer = testing.sql_layer


class TestModelUser(test_model_user.TestModelUser):
    layer = testing.sql_layer


class TestModelUsers(test_model_users.TestModelUsers):
    layer = testing.sql_layer


class TestBrowserActions(test_browser_actions.TestBrowserActions):
    layer = testing.sql_layer


class TestBrowserAutoincrement(test_browser_autoincrement.TestBrowserAutoincrement):
    layer = testing.sql_layer


class TestBrowserExpires(test_browser_expires.TestBrowserExpires):
    layer = testing.sql_layer


class TestBrowserGroup(test_browser_group.TestBrowserGroup):
    layer = testing.sql_layer


class TestBrowserGroups(test_browser_groups.TestBrowserGroups):
    layer = testing.sql_layer


class TestBrowserPortrait(test_browser_portrait.TestBrowserPortrait):
    layer = testing.sql_layer


class TestBrowserPrincipal(test_browser_principal.TestBrowserPrincipal):
    layer = testing.sql_layer


class TestBrowserRemote(test_browser_remote.TestBrowserRemote):
    layer = testing.sql_layer


class TestBrowserRoles(test_browser_roles.TestBrowserRoles):
    layer = testing.sql_layer


class TestBrowserRoot(test_browser_root.TestBrowserRoot):
    layer = testing.sql_layer


class TestBrowserSettings(test_browser_settings.TestBrowserSettings):
    layer = testing.sql_layer


class TestBrowserUser(test_browser_user.TestBrowserUser):
    layer = testing.sql_layer


class TestBroeserUsers(test_browser_users.TestBrowserUsers):
    layer = testing.sql_layer


class TestBrowserUtils(test_browser_utils.TestBrowserUtils):
    layer = testing.sql_layer



from cone.app import get_root
from cone.app import main_hook
from cone.app import ugm_backend
from cone.app.ugm import UGMFactory
from sqlalchemy import engine_from_config
from sqlalchemy import MetaData
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker
from zope.sqlalchemy import register
import os


###############################################################################
# Utils
###############################################################################

# key used for storing SQL session on request environment
session_key = 'cone.sql.session'


def get_session(request):
    """Return request related SQL session.
    """
    return request.environ[session_key]


# session setup handler registry
_session_setup_handlers = list()


def sql_session_setup(ob):
    """Decorator for registering SQL session setup handlers.

    A function decorated with ``sql_session_setup`` must accept a ``session``
    argument and is supposed to register SQLAlchemy event handlers and other
    setup tasks in conjunction with the created SQL session.
    """
    _session_setup_handlers.append(ob)
    return ob


def setup_session(session):
    """Call all registered session setup handlers.
    """
    for session_setup_callback in _session_setup_handlers:
        session_setup_callback(session)


def use_tm():
    """Flag whether transaction manager is used.
    """
    return os.environ.get('CONE_SQL_USE_TM') == '1'


###############################################################################
# DB initialization
###############################################################################

metadata = MetaData()
SQLBase = declarative_base(metadata=metadata)
DBSession = scoped_session(sessionmaker())


def initialize_sql(engine):
    """Basic SQL initialization.
    """
    DBSession.configure(bind=engine)
    metadata.bind = engine
    metadata.create_all(engine)


###############################################################################
# Session Factory
###############################################################################

class SQLSessionFactory(object):
    """SQL session factory.
    """

    def __init__(self, settings, prefix):
        self.engine = engine_from_config(settings, prefix=prefix)
        self.maker = sessionmaker(bind=self.engine)

    def __call__(self):
        session = self.maker()
        setup_session(session)
        return session


# Global session factory singleton.
session_factory = None


###############################################################################
# WSGI
###############################################################################

class WSGISQLSession(object):
    """WSGI framework component that opens and closes a SQL session.

    Downstream applications will have the session in the environment,
    normally under the key 'cone.sql.session'.
    """

    def __init__(self, next_app, session_key=session_key):
        self.next_app = next_app
        self.session_key = session_key

    def __call__(self, environ, start_response):
        session = session_factory()
        register(session)
        environ[self.session_key] = session
        try:
            result = self.next_app(environ, start_response)
            return result
        finally:
            session.close()


def make_app(next_app, global_conf, **local_conf):
    """Create ``WSGISQLSession``.
    """
    from cone import sql
    sql.session_key = local_conf.get('session_key', sql.session_key)
    return WSGISQLSession(next_app, sql.session_key)


###############################################################################
# Cone startup integration
###############################################################################

@main_hook
def initialize_cone_sql(config, global_config, settings):
    """Cone startup application initialization.
    """
    # database initialization
    prefix = 'sql.db.'
    if settings.get('{}url'.format(prefix), None) is None:  # pragma: no cover
        return
    if settings.get('ugm.backend') == 'sql':
        # If SQL configured as UGM backend, import ugm module to ensure proper
        # table creation at initialize_sql time.
        import cone.sql.ugm  # noqa
    global session_factory
    session_factory = SQLSessionFactory(settings, prefix)
    initialize_sql(session_factory.engine)
    use_tm = settings.get('pyramid.includes', '').find('pyramid_tm') > -1
    os.environ['CONE_SQL_USE_TM'] = '1' if use_tm else '0'


###############################################################################
# UGM factory
###############################################################################

@ugm_backend('sql')
class SqlUGMFactory(UGMFactory):
    """UGM backend factory for SQL based UGM implementation.
    """

    def __init__(self, settings):
        # cone.ugm active, users and groups attributes are read from ugm config
        if settings.get('cone.plugins').find('cone.ugm') > -1:
            from cone.ugm.utils import general_settings
            model = get_root()
            ugm_settings = general_settings(model).attrs
            self.user_attrs = ugm_settings.users_form_attrmap.keys()
            self.group_attrs = ugm_settings.groups_form_attrmap.keys()
        # users and groups attributes are read from application ini file
        else:
            self.user_atts = [
                attr.strip() for attr in
                settings.get('sql.user_attrs', '').split(',')
                if attr.strip()
            ]
            self.group_attrs = [
                attr.strip() for attr in
                settings.get('sql.group_attrs', '').split(',')
                if attr.strip()
            ]
        self.binary_attrs = [
            attr.strip() for attr in
            settings.get('sql.binary_attrs', '').split(',')
            if attr.strip()
        ]
        self.log_auth = settings.get('sql.log_auth') in ['true', 'True', '1']

    def __call__(self):
        from cone.sql.ugm import Ugm
        return Ugm(
            name="sql_ugm",
            parent=None,
            user_attrs=self.user_attrs,
            group_attrs=self.group_attrs,
            binary_attrs=self.binary_attrs,
            log_auth=self.log_auth
        )



from cone.app import get_root
from cone.app import register_entry
from cone.sql import get_session
from cone.sql import SQLBase
from cone.sql import testing
from cone.sql import use_tm
from cone.sql.model import GUID
from cone.sql.model import SQLRowNode
from cone.sql.model import SQLTableNode
from cone.sql.model import UNICODE_TYPE
from node.tests import NodeTestCase
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.dialects.postgresql.base import UUID
from sqlalchemy.engine import default
from sqlalchemy.orm.session import Session
from sqlalchemy.orm.unitofwork import UOWTransaction
from sqlalchemy.sql.sqltypes import CHAR
import os
import uuid


def reset_entry_registry(fn):
    """Decorator for tests using entry node registry
    """
    def wrapper(*a, **kw):
        root = get_root()
        factories = set(root.factories.keys())
        try:
            fn(*a, **kw)
        finally:
            for key in set(root.factories.keys()).difference(factories):
                del root.factories[key]
    return wrapper


class UUIDAsPrimaryKeyRecord(SQLBase):
    """Record with UUID as primary key.
    """
    __tablename__ = 'uuid_as_primary_key'
    uid_key = Column(GUID, primary_key=True)
    field = Column(String)


class UUIDAsKeyNode(SQLRowNode):
    record_class = UUIDAsPrimaryKeyRecord


class UUIDAsKeyContainer(SQLTableNode):
    record_class = UUIDAsPrimaryKeyRecord
    child_factory = UUIDAsKeyNode


class StringAsPrimaryKeyRecord(SQLBase):
    """Record with string as primary key.
    """
    __tablename__ = 'string_as_primary_key'
    string_key = Column(String, primary_key=True)
    field = Column(String)


class StringAsKeyNode(SQLRowNode):
    record_class = StringAsPrimaryKeyRecord


class StringAsKeyContainer(SQLTableNode):
    record_class = StringAsPrimaryKeyRecord
    child_factory = StringAsKeyNode


class IntegerAsPrimaryKeyRecord(SQLBase):
    """Record with integer as primary key.
    """
    __tablename__ = 'integer_as_primary_key'
    integer_key = Column(Integer, primary_key=True)
    field = Column(String)


class IntegerAsKeyNode(SQLRowNode):
    record_class = IntegerAsPrimaryKeyRecord


class IntegerAsKeyContainer(SQLTableNode):
    record_class = IntegerAsPrimaryKeyRecord
    child_factory = IntegerAsKeyNode


class TestModel(NodeTestCase):
    layer = testing.sql_layer

    def test_GUID(self):
        # Platform independent GUID data type

        # Define a dummy dialect
        class DummyDialect(default.DefaultDialect):
            name = None

        dialect = DummyDialect()

        # Instanciate ``GUID`` data type
        guid = GUID()

        # Test ``load_dialect_impl``
        dialect.name = 'postgresql'
        res = guid.load_dialect_impl(dialect)
        self.assertTrue(isinstance(res, UUID))

        dialect.name = 'other'
        res = guid.load_dialect_impl(dialect)
        self.assertTrue(isinstance(res, CHAR))

        # Test ``process_bind_param``
        dialect.name = 'postgresql'
        guid.process_bind_param(None, dialect)

        value = uuid.UUID('d8f1d964-9f2f-4df5-9f30-c5a90052576d')
        self.assertEqual(
            guid.process_bind_param(value, dialect),
            'd8f1d964-9f2f-4df5-9f30-c5a90052576d'
        )

        dialect.name = 'other'
        self.assertEqual(
            guid.process_bind_param(value, dialect),
            'd8f1d9649f2f4df59f30c5a90052576d'
        )

        value = str(value)
        self.assertEqual(
            guid.process_bind_param(value, dialect),
            'd8f1d9649f2f4df59f30c5a90052576d'
        )

        # Test ``process_result_value``
        guid.process_result_value(None, dialect)

        self.assertEqual(
            guid.process_result_value(value, dialect),
            uuid.UUID('d8f1d964-9f2f-4df5-9f30-c5a90052576d')
        )

    @reset_entry_registry
    def test_UUID_as_primary_key(self):
        # Resgister entry
        register_entry('uuid_as_key_container', UUIDAsKeyContainer)

        # Get container from root
        root = get_root()
        container = root['uuid_as_key_container']

        # Add node to container
        node_uid = '6090411e-d249-4dc6-9da1-74172919f1ed'
        node = container[node_uid] = UUIDAsKeyNode()
        node.attrs['field'] = u'Value'

        # Persist data
        container()

        # Query data record using SQLAlchemy directly
        request = self.layer.new_request()
        session = get_session(request)
        rec = session.query(UUIDAsPrimaryKeyRecord).get(uuid.UUID(node_uid))
        self.assertTrue(isinstance(rec, UUIDAsPrimaryKeyRecord))

        # Get children via node API
        node = container['6090411e-d249-4dc6-9da1-74172919f1ed']
        self.assertTrue(isinstance(node, UUIDAsKeyNode))
        self.assertEqual(
            container.keys(),
            ['6090411e-d249-4dc6-9da1-74172919f1ed']
        )
        self.assertEqual(len(container.values()), 1)
        self.assertEqual(
            container.values()[0].name,
            '6090411e-d249-4dc6-9da1-74172919f1ed'
        )
        self.assertEqual(len(container.items()), 1)
        self.assertEqual(
            container.items()[0][0],
            '6090411e-d249-4dc6-9da1-74172919f1ed'
        )
        self.assertEqual(
            container.items()[0][1].name,
            '6090411e-d249-4dc6-9da1-74172919f1ed'
        )
        self.assertEqual(
            sorted(node.attrs.items()),
            [
                ('field', u'Value'),
                ('uid_key', uuid.UUID('6090411e-d249-4dc6-9da1-74172919f1ed'))
            ]
        )

    @reset_entry_registry
    def test_string_as_primary_key(self):
        # Resgister entry.
        register_entry('string_as_key_container', StringAsKeyContainer)

        # Get container from root
        root = get_root()
        container = root['string_as_key_container']

        # Add node to container
        node = container[u'key'] = StringAsKeyNode()
        node.attrs['field'] = u'Value'

        # Persist data
        container()

        # Query data record using SQLAlchemy directly
        request = self.layer.new_request()
        session = get_session(request)
        rec = session.query(StringAsPrimaryKeyRecord).get(u'key')
        self.assertTrue(isinstance(rec, StringAsPrimaryKeyRecord))

        # Get children via node API
        node = container['key']
        self.assertTrue(isinstance(node, StringAsKeyNode))
        self.assertEqual(container.keys(), ['key'])
        self.assertEqual(len(container.values()), 1)
        self.assertEqual(container.values()[0].name, 'key')
        self.assertEqual(len(container.items()), 1)
        self.assertEqual(container.items()[0][0], 'key')
        self.assertEqual(container.items()[0][1].name, 'key')
        self.assertEqual(
            sorted(node.attrs.items()),
            [('field', u'Value'), ('string_key', u'key')]
        )

    @reset_entry_registry
    @testing.delete_table_records(IntegerAsPrimaryKeyRecord)
    def test_int_as_primary_key(self):
        # Resgister entry
        register_entry('integer_as_key_container', IntegerAsKeyContainer)

        # Get container from root
        root = get_root()
        container = root['integer_as_key_container']

        # Add node to container
        node = container['1234'] = IntegerAsKeyNode()
        node.attrs['field'] = u'Value'

        # Persist data
        container()

        # Query data record using SQLAlchemy directly
        request = self.layer.new_request()
        session = get_session(request)
        rec = session.query(IntegerAsPrimaryKeyRecord).get('1234')
        self.assertTrue(isinstance(rec, IntegerAsPrimaryKeyRecord))

        # Get children via node API
        node = container['1234']
        self.assertTrue(isinstance(node, IntegerAsKeyNode))
        self.assertEqual(container.keys(), ['1234'])
        self.assertEqual(len(container.values()), 1)
        self.assertEqual(container.values()[0].name, '1234')
        self.assertEqual(len(container.items()), 1)
        self.assertEqual(container.items()[0][0], '1234')
        self.assertEqual(container.items()[0][1].name, '1234')
        self.assertEqual(
            sorted(node.attrs.items()),
            [('field', u'Value'), ('integer_key', 1234)]
        )

    def test_data_type_converters(self):
        # SQLAlchemy data types for primary keys can be extended on
        # ``data_type_converters``
        converters = sorted(
            SQLTableNode.data_type_converters.items(),
            key=lambda x: x[0].__name__
        )
        self.assertEqual(converters, [
            (GUID, uuid.UUID),
            (Integer, int),
            (String, UNICODE_TYPE)
        ])

    @reset_entry_registry
    @testing.delete_table_records(IntegerAsPrimaryKeyRecord)
    def test_node_api(self):
        # Resgister entry
        register_entry('integer_as_key_container', IntegerAsKeyContainer)

        # ``__getitem__`` and ``__setitem__`` raise a ``KeyError`` if node name
        # cannot be converted to primary key data type
        root = get_root()
        container = root['integer_as_key_container']

        err = self.expect_error(KeyError, container.__getitem__, 'a')
        expected = (
            '"Failed to convert node name to expected primary key data type: '
            'invalid literal for int() with base 10: \'a\'"'
        )
        self.assertEqual(str(err), expected)

        err = self.expect_error(
            KeyError,
            container.__setitem__,
            'a',
            IntegerAsKeyNode()
        )
        expected = (
            '"Failed to convert node name to expected primary key data type: '
            'invalid literal for int() with base 10: \'a\'"'
        )
        self.assertEqual(str(err), expected)

        # If primary key attribute is set on node and given name on
        # ``__setitem__`` not matches attribute value, a ``KeyError`` is thrown.
        child = IntegerAsKeyNode()
        child.attrs['integer_key'] = 123
        err = self.expect_error(
            KeyError,
            container.__setitem__,
            '124',
            child
        )
        expected = "'Node name must match primary key attribute value: 124 != 123'"
        self.assertEqual(str(err), expected)

        # Access inexistent child
        err = self.expect_error(
            KeyError,
            container.__getitem__,
            '124'
        )
        self.assertEqual(str(err), "'124'")

        # If primary key attribute not set, it gets automatically set by name
        # on ``__setitem__``.
        child = IntegerAsKeyNode()
        container['123'] = child
        self.assertEqual(
            sorted(child.attrs.items()),
            [('field', None), ('integer_key', 123)]
        )

        # SQL model column values can be accessed and set via ``attrs``
        child.attrs['field'] = u'Value'
        self.assertEqual(
            sorted(child.attrs.items()),
            [('field', u'Value'), ('integer_key', 123)]
        )

        # SQL model gets persisted on ``__call__``.
        container()

        request = self.layer.new_request()
        session = get_session(request)
        res = session.query(IntegerAsPrimaryKeyRecord).all()
        self.assertEqual(len(res), 1)

        # Override child
        child = IntegerAsKeyNode()
        child.attrs['field'] = u'Other Value'
        container['123'] = child
        self.assertEqual(
            sorted(child.attrs.items()),
            [('field', u'Other Value'), ('integer_key', 123)]
        )

        container()
        request = self.layer.new_request()
        session = get_session(request)
        res = session.query(IntegerAsPrimaryKeyRecord).all()
        self.assertEqual(len(res), 1)

        # Delete child
        del container['123']

        request = self.layer.new_request()
        session = get_session(request)
        res = session.query(IntegerAsPrimaryKeyRecord).all()
        self.assertEqual(len(res), 0)

        # Other than most other node implementations, ``TableRowNodes`` can be
        # persisted without being hooked up to the tree directly.
        child = IntegerAsKeyNode()
        child.attrs['integer_key'] = 1234
        child.attrs['field'] = u'Value'
        child()

        self.assertEqual(container.keys(), ['1234'])

        # Update Child
        child = container['1234']
        child.attrs['field'] = u'Updated Value'
        child()

        request = self.layer.new_request()
        session = get_session(request)
        self.assertEqual(
            session.query(IntegerAsPrimaryKeyRecord).first().field,
            u'Updated Value'
        )

        # Access inexisting attributes
        err = self.expect_error(
            KeyError,
            child.attrs.__getitem__,
            'inexistent'
        )
        self.assertEqual(str(err), "'Unknown attribute: inexistent'")

        err = self.expect_error(
            KeyError,
            child.attrs.__setitem__,
            'inexistent',
            'Value'
        )
        self.assertEqual(str(err), "'Unknown attribute: inexistent'")

        # SQL row node attributes cannot be deleted
        err = self.expect_error(
            KeyError,
            child.attrs.__delitem__,
            'field'
        )
        self.assertEqual(str(err), "'Deleting of attributes not allowed'")

        # SQL row node is a leaf thus containment API always raises KeyError
        # and iter returns empty result.
        err = self.expect_error(
            KeyError,
            child.__setitem__,
            'foo',
            'foo'
        )
        self.assertEqual(str(err), "'foo'")

        err = self.expect_error(
            KeyError,
            child.__getitem__,
            'foo'
        )
        self.assertEqual(str(err), "'foo'")

        err = self.expect_error(
            KeyError,
            child.__delitem__,
            'foo'
        )
        self.assertEqual(str(err), "'foo'")

        self.assertEqual(list(iter(child)), [])

    @reset_entry_registry
    @testing.delete_table_records(IntegerAsPrimaryKeyRecord)
    def test_use_tm(self):
        # Resgister entry
        register_entry('integer_as_key_container', IntegerAsKeyContainer)

        root = get_root()
        container = root['integer_as_key_container']

        # create a base entry for direct node modification test
        request = self.layer.new_request()
        node = container['1'] = IntegerAsKeyNode()
        node.attrs['field'] = 'Value'
        node()

        # transaction manager used, calling nodes flushes session
        os.environ['CONE_SQL_USE_TM'] = '1'
        self.assertTrue(use_tm())

        # modify existing node
        node.attrs['field'] = 'New'
        node()

        # create new node
        container['2'] = IntegerAsKeyNode()
        container()

        session = get_session(request)
        res = session.query(IntegerAsPrimaryKeyRecord).all()
        self.assertEqual(len(res), 2)
        self.assertEqual(res[0].field, 'New')

        # rollback works, session was just flushed
        session.rollback()
        res = session.query(IntegerAsPrimaryKeyRecord).all()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0].field, 'Value')

        # no transaction manager used, calling nodes commits session
        os.environ['CONE_SQL_USE_TM'] = '0'
        self.assertFalse(use_tm())

        # modify existing node
        node.attrs['field'] = 'New'
        node()

        # create new node
        container['2'] = IntegerAsKeyNode()
        container()

        session = get_session(request)
        res = session.query(IntegerAsPrimaryKeyRecord).all()
        self.assertEqual(len(res), 2)
        self.assertEqual(res[0].field, 'New')

        # rollback has not effect, session was commited
        session.rollback()
        res = session.query(IntegerAsPrimaryKeyRecord).all()
        self.assertEqual(len(res), 2)
        self.assertEqual(res[0].field, 'New')

    @reset_entry_registry
    @testing.delete_table_records(IntegerAsPrimaryKeyRecord)
    def test_sql_session_setup(self):
        # Resgister entry
        register_entry('integer_as_key_container', IntegerAsKeyContainer)

        # Test ``sql_session_setup``. The SQL session setup handler is defined
        # in ``cone.sql.testing`` and registers a callback to ``after_flush``
        # event. Patch desired callback reference and test whether it's called.
        class Callback(object):
            session = None
            flush_context = None

            def __call__(self, session, flush_context):
                self.session = session
                self.flush_context = flush_context

        callback = Callback()
        testing.test_after_flush = callback

        root = get_root()
        container = root['integer_as_key_container']
        node = container['1236'] = IntegerAsKeyNode()
        node.attrs['field'] = u'Value'
        container()

        self.assertTrue(isinstance(callback.session, Session))
        self.assertTrue(isinstance(callback.flush_context, UOWTransaction))

        testing.test_after_flush = None



from cone.sql import testing
from cone.sql.ugm import Base
from cone.sql.ugm import Group
from cone.sql.ugm import SQLGroup
from cone.sql.ugm import SQLGroupAssignment
from cone.sql.ugm import SQLPrincipal
from cone.sql.ugm import SQLUser
from cone.sql.ugm import Ugm
from cone.sql.ugm import User
from node.tests import NodeTestCase
from sqlalchemy.engine import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.attributes import flag_modified
import os
import unittest


def temp_database(fn):
    """This decorator creates an in-memory sqlite db for testing the user
    classes.
    """

    def wrapper(self):
        curdir = os.path.dirname(__file__)
        fname = '%s/test.db' % curdir
        if os.path.exists(fname):
            os.remove(fname)
        uri = 'sqlite:///' + fname
        engine = create_engine(uri)
        Base.metadata.create_all(engine)
        sm = sessionmaker(bind=engine)
        session = sm()
        fn(self, session)

    return wrapper


class TestSqlModel(unittest.TestCase):

    @temp_database
    def test_sql_model(self, session):
        for name in ['phil', 'donald', 'dagobert', 'daisy']:
            session.add(SQLUser(id=name))
        session.flush()

        users = session.query(SQLUser).all()
        self.assertEqual(
            sorted([u.id for u in users]),
            ['dagobert', 'daisy', 'donald', 'phil']
        )

        for group in ['admins', 'members', 'editors', 'phil']:
            session.add(SQLGroup(id=group))
        session.flush()

        phil = session.query(SQLUser).filter(SQLUser.id == 'phil').one()
        donald = session.query(SQLUser).filter(SQLUser.id == 'donald').one()
        admins = session.query(SQLGroup).filter(SQLGroup.id == 'admins').one()
        members = session.query(SQLGroup).filter(SQLGroup.id == 'members').one()

        phil.principal_roles = ['manager', 'member']
        phil.groups.append(admins)
        phil.groups.append(members)

        donald.groups.append(members)

        session.flush()

        phil1 = session.query(SQLUser).filter(SQLUser.id == 'phil').one()
        self.assertEqual(sorted(phil1.principal_roles), ['manager', 'member'])
        self.assertTrue(admins in phil1.groups)
        self.assertTrue(members in phil1.groups)

        members = session.query(SQLGroup).filter(SQLGroup.id == 'members').one()
        phil_group = session.query(SQLGroup).filter(SQLGroup.id == 'phil').one()
        phil_group.frunz = 42

        self.assertEqual(phil_group.id, phil.id)
        self.assertNotEqual(phil_group.guid, phil.guid)

        self.assertTrue(phil in members.users)
        self.assertTrue(donald in members.users)

        phil_group.users.append(phil)
        self.assertTrue(phil_group in phil.groups)

        session.commit()


class TestSqlUgm(NodeTestCase):
    layer = testing.sql_layer

    @testing.delete_table_records(SQLPrincipal)
    @testing.delete_table_records(SQLGroup)
    @testing.delete_table_records(SQLGroupAssignment)
    @testing.delete_table_records(SQLUser)
    def test_ugm(self):
        os.environ['CONE_SQL_USE_TM'] = '0'
        self.layer.new_request()

        ugm = Ugm(
            name='sql_ugm',
            parent=None,
            user_attrs=['phone', 'address'],
            group_attrs=['description'],
            binary_attrs=[],
            log_auth=True
        )
        users = ugm.users
        groups = ugm.groups

        self.assertTrue(users.ugm is ugm)
        self.assertTrue(groups.ugm is ugm)

        # create some users with attribute
        ids = ['phil', 'donald', 'dagobert', 'mickey']
        for count, id in enumerate(ids):
            users.create(
                id,
                height=count + 1,
                email='%s@bluedynamics.net' % id,
                status='super%s' % (count + 1)
            )

        self.assertTrue('phil' in users)
        self.assertFalse('zworkb' in users)

        # give phil a password
        self.assertEqual(users['phil'].record.password, None)
        users.set_hashed_pw('phil', users.hash_passwd('test123'))
        self.assertNotEqual(users['phil'].record.password, None)

        self.assertEqual(users['donald'].record.password, None)
        users['donald'].passwd(None, 'test123')
        self.assertNotEqual(users['donald'].record.password, None)

        self.assertEqual(users['dagobert'].record.password, None)
        users.passwd('dagobert', None, 'test124')
        self.assertNotEqual(users['dagobert'].record.password, None)

        self.assertFalse(users.authenticate('zworkb', 'test123'))
        self.assertFalse(users.authenticate(None, "foo"))
        self.assertFalse(users.authenticate("foo", None))
        self.assertFalse(users.authenticate(None, None))

        self.assertTrue(users['phil'].authenticate('test123'))
        self.assertTrue(users.authenticate('phil', 'test123'))
        self.assertTrue(users.authenticate('donald', 'test123'))
        self.assertTrue(users.authenticate('dagobert', 'test124'))

        # check user attributes
        self.assertEqual(users['phil'].record.data['height'], 1)
        self.assertEqual(users['donald'].record.data['height'], 2)
        self.assertEqual(users['phil'].record.data['status'], 'super1')
        self.assertEqual(users['donald'].record.data['status'], 'super2')

        # check __iter__
        ids1 = list(users)
        self.assertEqual(sorted(ids), sorted(ids1))

        # check login attribute (lets take email)
        # schlumpf and schlumpfine with 2 different login fields
        users.create('schlumpf', email='schlumpf@bluedynamics.net', login='email')
        users.create('schlumpfine', nickname='schlumpfinchen', login='nickname')

        schlumpfid = users.id_for_login('schlumpf@bluedynamics.net')
        self.assertEqual(schlumpfid, 'schlumpf')

        schlumpfineid = users.id_for_login('schlumpfinchen')
        self.assertEqual(schlumpfineid, 'schlumpfine')

        users.session.commit()

        # Test groups
        managers = groups.create('managers', title='Masters of the Universe')
        members = groups.create('members', title='the normal ones')
        managers1 = groups['managers']

        self.assertEqual(
            managers1.record.data['title'],
            'Masters of the Universe'
        )
        self.assertTrue(groups.ugm is ugm)

        managers.add('phil')

        for id in ids:
            members.add(id)

        users.session.commit()

        self.assertEqual(set(ids), set(list(members)))
        self.assertEqual(sorted(members.member_ids), sorted(ids))
        self.assertEqual(managers.member_ids, ['phil'])

        phil2 = managers['phil']
        self.assertIsInstance(phil2, User)

        # non group members should raise a KeyError
        self.assertRaises(KeyError, managers.__getitem__, 'donald')

        # Test roles

        # roles for a user
        users['phil'].add_role('Editor')
        users['phil'].add_role('Spam')
        self.assertEqual(sorted(users['phil'].roles), ['Editor', 'Spam'])

        users['phil'].remove_role('Spam')
        self.assertEqual(users['phil'].roles, ['Editor'])

        users.session.commit()

        # removing non-existing roles is tolerated
        users['phil'].remove_role('Spam')

        # roles for group
        groups['managers'].add_role('Manager')
        self.assertEqual(groups['managers'].roles, ['Manager'])

        groups['members'].add_role('Member')
        self.assertEqual(groups['members'].roles, ['Member'])

        # cumulative roles for the user -> user has all roles by his groups
        self.assertEqual(
            sorted(users['phil'].roles),
            ['Editor', 'Manager', 'Member']
        )

        # get group instances of a user
        self.assertEqual(len(users['phil'].groups), 2)
        for g in users['phil'].groups:
            self.assertIsInstance(g, Group)

        # group_ids shall match group instances
        self.assertEqual(
            set(users['phil'].group_ids),
            set([g.id for g in users['phil'].groups])
        )

        # delete a group membership
        del managers['phil']
        users.session.commit()  # needs commit, flush() is not sufficient

        self.assertTrue(users['phil'] not in managers.users)  # needs refresh
        self.assertTrue(users['phil'] not in groups['managers'].users)
        self.assertTrue('phil' not in managers.member_ids)
        self.assertTrue('managers' not in users['phil'].group_ids)
        self.assertTrue(groups['managers'] not in users['phil'].groups)

        # ugm-level role management
        ugm.add_role('Snuff', users['phil'])
        self.assertTrue('Snuff' in users['phil'].roles)

        ugm.add_role('Smurf', groups['managers'])
        self.assertTrue('Smurf' in groups['managers'].roles)

        ugm.remove_role('Snuff', users['phil'])
        self.assertFalse('Snuff' in users['phil'].roles)

        ugm.remove_role('Smurf', groups['managers'])
        self.assertFalse('Smurf' in groups['managers'].roles)

        # searching

        # search by int
        r1 = users.search(criteria=dict(height=1))
        self.assertEqual(len(r1), 1)
        self.assertEqual(r1[0], 'phil')

        # search by string
        r2 = users.search(criteria=dict(status='super1'))
        self.assertEqual(len(r2), 1)
        self.assertEqual(r2[0], 'phil')

        # search with or
        r3 = users.search(
            criteria=dict(
                status='super1',
                height=2
            ),
            or_search=True
        )
        self.assertEqual(len(r3), 2)
        self.assertEqual(set(r3), set(('phil', 'donald',)))

        # search with wildcards in dynamic fields
        r4 = users.search(criteria=dict(status='super*'), exact_match=False)
        self.assertEqual(len(r4), 4)
        self.assertEqual(set(r4), set(('phil', 'donald', 'dagobert', 'mickey')))

        r5 = users.search(criteria=dict(id='d*'), exact_match=False)
        self.assertEqual(len(r5), 2)
        self.assertEqual(set(r5), set(('donald', 'dagobert')))

        # blank search should return all entries
        r6 = users.search()
        self.assertEqual(len(r6), 6)  # simply all of them

        # search for login field
        r7 = users.search(criteria=dict(login='nickname'))
        self.assertEqual(set(r7), {'schlumpfine'})

        # search with attrlist
        r8 = users.search(attrlist=['login', 'height', 'status'])
        self.assertEqual(sorted(r8), sorted([
            ('donald', {'height': 2, 'login': None, 'status': 'super2'}),
            ('dagobert', {'height': 3, 'login': None, 'status': 'super3'}),
            ('mickey', {'height': 4, 'login': None, 'status': 'super4'}),
            ('schlumpf', {'height': None, 'login': 'email', 'status': None}),
            ('schlumpfine', {'height': None, 'login': 'nickname', 'status': None}),
            ('phil', {'height': 1, 'login': None, 'status': 'super1'})])
        )

        r9 = users.search(attrlist=[])
        self.assertEqual(sorted(r9), sorted([
            ('donald', {
                'email': 'donald@bluedynamics.net',
                'height': 2,
                'login': None,
                'status': 'super2'
            }), ('dagobert', {
                'email': 'dagobert@bluedynamics.net',
                'height': 3,
                'login': None,
                'status': 'super3'
            }), ('mickey', {
                'email': 'mickey@bluedynamics.net',
                'height': 4,
                'login': None,
                'status': 'super4'
            }), ('schlumpf', {
                'email': 'schlumpf@bluedynamics.net',
                'login': 'email'
            }), ('schlumpfine', {
                'login': 'nickname',
                'nickname': 'schlumpfinchen'
            }), ('phil', {
                'email': 'phil@bluedynamics.net',
                'height': 1,
                'login': None,
                'status': 'super1'
            })
        ]))

        # Exact searches with empty result shall throw up
        self.assertRaises(ValueError, lambda: users.search(
            exact_match=True,
            criteria=dict(
                id='unobtainable'
            )
        ))

        r10 = users.search(
            exact_match=False,
            criteria=dict(
                id='unobtainable'
            )
        )
        self.assertEqual(list(r10), [])

        r11 = users.search(
            exact_match=False,
            criteria=dict(
                id='unobtainable'
            ),
            or_search=True
        )
        self.assertEqual(list(r11), [])

        # shall work with groups too
        gids = groups.search(criteria=dict(title='Masters*'))
        self.assertEqual(gids, ['managers'])

        self.assertEqual(len(groups.keys()), 2)
        gids = groups.search(criteria=dict(id='*foo*'), or_search=True)
        self.assertEqual(len(gids), 0)

        del os.environ['CONE_SQL_USE_TM']

        # change fields of a user
        ugm.session.commit()

        donald = users['donald']
        donald.record.login = 'mail'
        donald.record.data['mail'] = 'donald@duck.com'

        flag_modified(donald.record, 'data')
        ugm.session.commit()

        donald1 = users['donald']
        self.assertEqual(donald1.record.login, 'mail')
        self.assertEqual(donald1.record.data['mail'], 'donald@duck.com')



from cone.app.security import PrincipalACL
from cone.sql import get_session
from cone.sql import SQLBase
from cone.sql.model import GUID
from node.behaviors import Adopt
from node.behaviors import DefaultInit
from node.behaviors import NodeChildValidate
from node.behaviors import Nodify
from node.interfaces import IUUID
from node.utils import instance_property
from plumber import default
from plumber import plumbing
from pyramid.threadlocal import get_current_request
from sqlalchemy import and_
from sqlalchemy import Column
from sqlalchemy import String
import uuid


class PrincipalRoleRecord(SQLBase):
    __tablename__ = 'principal_roles'

    rec_id = Column(GUID, primary_key=True, default=lambda: uuid.uuid4())
    node_id = Column(GUID)
    principal_id = Column(String)
    role = Column(String)


@plumbing(
    NodeChildValidate,
    Adopt,
    DefaultInit,
    Nodify)
class SQLPrincipalRoles(object):
    """Principal roles from sql.
    """
    allow_non_node_childs = True

    @property
    def session(self):
        return get_session(get_current_request())

    def _roles_for(self, principal_id):
        res = self.session\
            .query(PrincipalRoleRecord.role)\
            .filter(and_(
                PrincipalRoleRecord.node_id == self.parent.uuid,
                PrincipalRoleRecord.principal_id == principal_id
            ))\
            .distinct()
        return [rec.role for rec in res]

    def __getitem__(self, name):
        return self._roles_for(name)

    def __setitem__(self, name, value):
        session = self.session
        existing = self._roles_for(name)
        for role in value:
            if role not in existing:
                session.add(PrincipalRoleRecord(
                    node_id=self.parent.uuid,
                    principal_id=name,
                    role=role
                ))
        delete = list()
        for role in existing:
            if role not in value:
                delete.append(role)
        if delete:
            res = session\
                .query(PrincipalRoleRecord)\
                .filter(and_(
                    PrincipalRoleRecord.node_id == self.parent.uuid,
                    PrincipalRoleRecord.principal_id == name,
                    PrincipalRoleRecord.role.in_(delete)
                ))\
                .all()
            for record in res:
                session.delete(record)

    def __delitem__(self, name):
        session = self.session
        res = session\
            .query(PrincipalRoleRecord)\
            .filter(and_(
                PrincipalRoleRecord.node_id == self.parent.uuid,
                PrincipalRoleRecord.principal_id == name
            ))\
            .all()
        for record in res:
            session.delete(record)

    def __iter__(self):
        res = self.session\
            .query(PrincipalRoleRecord.principal_id)\
            .filter(PrincipalRoleRecord.node_id == self.parent.uuid)\
            .distinct()
        for rec in res:
            yield rec.principal_id


class SQLPrincipalACL(PrincipalACL):
    """Principal ACL stored in relational database.
    """

    @default
    @instance_property
    def principal_roles(self):
        if not IUUID.providedBy(self):
            raise RuntimeError(u"%s does not implement IUUID" % str(self))
        return SQLPrincipalRoles('principal_roles', parent=self)



from cone.sql import SQLBase as Base
from cone.sql import use_tm
from cone.sql.model import GUID
from cone.sql.model import SQLRowNodeAttributes
from cone.sql.model import SQLSession
from cone.sql.model import UNICODE_TYPE
from datetime import datetime
from node.behaviors import Adopt
from node.behaviors import Attributes
from node.behaviors import DefaultInit
from node.behaviors import NodeChildValidate
from node.behaviors import Nodespaces
from node.behaviors import Nodify
from node.ext.ugm import Group as BaseGroup
from node.ext.ugm import Groups as BaseGroups
from node.ext.ugm import Ugm as BaseUgm
from node.ext.ugm import User as BaseUser
from node.ext.ugm import Users as BaseUsers
from node.utils import UNSET
from operator import or_
from plumber import Behavior
from plumber import default
from plumber import override
from plumber import plumbing
from sqlalchemy import and_
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import inspect
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.dialects.sqlite.base import SQLiteTypeCompiler
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.orm.exc import NoResultFound
import base64
import hashlib
import itertools
import os
import uuid

# HACK: Force sqlite to alias JSONB as JSON. This allows to use JSONB for the
#       sqlalchemy variant, which is much more efficient when it comes to
#       indexing and searching.
SQLiteTypeCompiler.visit_JSONB = SQLiteTypeCompiler.visit_JSON


###############################################################################
# SQLAlchemy model classes
###############################################################################

class SQLPrincipal(Base):
    __tablename__ = 'principal'
    discriminator = Column(String)
    __mapper_args__ = {'polymorphic_on': discriminator}
    guid = Column(
        GUID,
        default=lambda: str(uuid.uuid4()),
        index=True,
        primary_key=True
    )
    data = Column(JSONB)
    principal_roles = Column(JSONB, default=[])
    created = Column(DateTime, default=datetime.now)

    def get_attribute(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            return self.data.get(key)


class SQLGroup(SQLPrincipal):
    __tablename__ = 'group'
    __mapper_args__ = {'polymorphic_identity': 'sqlgroup'}
    guid = Column(
        GUID,
        ForeignKey('principal.guid', deferrable=True),
        primary_key=True
    )
    id = Column(String, unique=True)
    users = association_proxy(
        'group_assignments',
        'users',
        creator=lambda c: SQLGroupAssignment(users=c)
    )
    group_assignments = relationship(
        'SQLGroupAssignment',
        backref='groups',
        primaryjoin='SQLGroupAssignment.groups_guid == SQLGroup.guid',
        cascade='save-update, merge, delete, delete-orphan'
    )


class SQLGroupAssignment(Base):
    __tablename__ = 'group_assignment'
    groups_guid = Column(
        GUID,
        ForeignKey('group.guid', deferrable=True),
        primary_key=True,
        nullable=False
    )
    users_guid = Column(
        GUID,
        ForeignKey('user.guid', deferrable=True),
        primary_key=True,
        nullable=False
    )


class SQLUser(SQLPrincipal):
    __tablename__ = 'user'
    __mapper_args__ = {'polymorphic_identity': 'sqluser'}
    guid = Column(
        GUID,
        ForeignKey('principal.guid', deferrable=True),
        primary_key=True
    )
    login = Column(String)
    id = Column(String, unique=True)
    password = Column(String)
    first_login = Column(DateTime, nullable=True)
    last_login = Column(DateTime, nullable=True)
    groups = association_proxy(
        'group_assignments',
        'groups',
        creator=lambda c: SQLGroupAssignment(groups=c)
    )
    group_assignments = relationship(
        'SQLGroupAssignment',
        backref='users',
        primaryjoin='SQLGroupAssignment.users_guid == SQLUser.guid',
        cascade='save-update, merge, delete, delete-orphan'
    )


###############################################################################
# Node classes
###############################################################################

class PrincipalAttributes(SQLRowNodeAttributes):
    """Consists of field configured in .ini file or ugm.xml and the given SQL
    schema fields.
    """

    @property
    def ugm(self):
        return self.parent.ugm

    def __setitem__(self, name, value):
        if value is UNSET:
            value = ''
        if value and name in self.binary_attrs:
            value = base64.b64encode(value).decode()
        if name in self.schema_attrs:
            setattr(self.record, name, value)
        else:
            self.record.data[name] = value
            flag_modified(self.record, 'data')

    def __getitem__(self, name):
        value = self.record.get_attribute(name)
        if value and name in self.binary_attrs:
            value = base64.b64decode(value)
        return value

    @property
    def _columns(self):
        if self.configured_attrs:
            return self.configured_attrs
        else:
            return self.inspected_attrs

    @property
    def schema_attrs(self):
        """Fields that are in the record schema without the technical fields.
        """
        tech_attrs = [
            'group_assignments', 'discriminator', 'guid',
            'data', 'principal_roles', 'password'
        ]
        schema_attrs = [
            f for f in
            inspect(self.record.__class__).attrs.keys()
            if f not in tech_attrs
        ]
        return schema_attrs

    @property
    def inspected_attrs(self):
        """Fields that are in the record schema + keys from record.data without
        the technical fields.
        """
        return self.schema_attrs + list(self.record.data.keys())

    @property
    def binary_attrs(self):
        return self.ugm.binary_attrs


class UserAttributes(PrincipalAttributes):

    @property
    def configured_attrs(self):
        return self.ugm.user_attrs


class GroupAttributeFactory(PrincipalAttributes):

    @property
    def configured_attrs(self):
        return self.ugm.group_attrs


class PrincipalBehavior(Behavior):

    record = default(None)
    """Reference to sqlalchemy record instance."""

    @override
    def __init__(self, parent, record):
        self.__parent__ = parent
        self.record = record

    @default
    @property
    def __name__(self):
        return self.id

    @default
    @property
    def ugm(self):
        return self.parent.parent

    @default
    @property
    def id(self):
        return self.record.id

    @default
    def add_role(self, role):
        if role not in self.own_roles:
            self.own_roles = self.own_roles + [role]

    @default
    def remove_role(self, role):
        if role in self.own_roles:
            # to trigger the json field
            self.own_roles = [r for r in self.own_roles if r != role]

    @property
    def own_roles(self):
        return self.record.principal_roles

    @default
    @own_roles.setter
    def own_roles(self, roles):
        self.record.principal_roles = roles

    @default
    @property
    def roles(self):
        return self.own_roles

    @default
    def __call__(self):
        if use_tm():
            self.session.flush()
        else:
            self.session.commit()


class UserBehavior(PrincipalBehavior, BaseUser):

    @default
    @property
    def group_ids(self):
        return [g.id for g in self.groups]

    @default
    @property
    def groups(self):
        return [self.ugm.groups[g.id] for g in self.record.groups]

    @default
    @property
    def roles(self):
        """Accumulate principal's roles + assigned groups' roles.
        """
        my_roles = self.own_roles
        all_roles = itertools.chain(
            my_roles,
            *[g.principal_roles for g in self.record.groups]
        )
        return list(set(all_roles))

    @default
    def authenticate(self, pw):
        return self.ugm.users.authenticate(self.id, pw)

    @default
    def passwd(self, old, new):
        return self.ugm.users.passwd(self.id, old, new)

    @default
    def attributes_factory(self, name, parent):
        return UserAttributes(name, parent, self.record)


@plumbing(
    UserBehavior,
    Nodespaces,
    Attributes,
    Nodify,
    Adopt,
    SQLSession)
class User(object):
    pass


class GroupBehavior(PrincipalBehavior, BaseGroup):

    @default
    @property
    def member_ids(self):
        return [u.id for u in self.record.users]

    @default
    def add(self, id):
        user = self.ugm.users[id]
        self.record.users.append(user.record)

    @default
    def __getitem__(self, key):
        res = self.ugm.users[key]

        if self.record not in res.record.groups:
            raise KeyError(key)

        return res

    @default
    def __delitem__(self, key):
        self.record.users.remove(self.ugm.users[key].record)

    @default
    def __iter__(self):
        # XXX: for groups with many many members this should be implemented lazy
        return iter(self.member_ids)

    @default
    @property
    def users(self):
        return [
            self.ugm.users[id]
            for id in self.member_ids
        ]

    @default
    def attributes_factory(self, name, parent):
        return GroupAttributeFactory(name, parent, self.record)


@plumbing(
    GroupBehavior,
    Nodespaces,
    Attributes,
    Nodify,
    SQLSession)
class Group(object):
    pass


class PrincipalsBehavior(Behavior):

    @default
    @property
    def ugm(self):
        return self.parent

    @override
    def search(self, criteria=None, attrlist=None,
               exact_match=False, or_search=False):
        typemap = {
            str: String,
            int: Integer
        }
        if criteria is None:
            criteria = {}

        op = or_ if or_search else and_
        cls = self.record_class
        fixed_attrs = ['id', 'login']
        fixed_attr_comparators = [
            getattr(cls, key) == criteria[key] if exact_match
            else getattr(cls, key).like(('%s' % criteria[key]).replace('*', '%%'))
            for key in fixed_attrs
            if key in criteria
        ]
        for key in fixed_attrs:
            criteria = criteria.copy()
            criteria.pop(key, None)

        def literal(value):
            lit = ('"%s"' % value) if isinstance(value, str) else value
            if exact_match:
                return lit
            else:
                return lit.replace('*', '%%') if isinstance(lit, str) else lit

        def field_selector(key, value):
            return cls.data[key].cast(String).cast(typemap[type(value)])

        def field_comparator(key, value):
            if not exact_match and isinstance(value, str):
                return field_selector(key, value).like(literal(value))
            else:
                return field_selector(key, value) == literal(value)

        dynamic_comparators = [
            field_comparator(key, value)
            for (key, value) in criteria.items()
        ]

        comparators = fixed_attr_comparators + dynamic_comparators
        if len(comparators) >= 2:
            clause = op(*comparators)
        elif len(comparators) == 1:
            clause = comparators[0]
        else:
            clause = None

        basequery = self.ugm.users.session.query(cls)
        if clause is not None:
            query = basequery.filter(clause)
        else:
            query = basequery

        binary_attrs = self.ugm.binary_attrs

        def get_attribute(p, k):
            value = p.get_attribute(k)
            if value and k in binary_attrs:
                value = base64.b64decode(value)
            return value

        # XXX: should we be lazy here and yield?, would be nice for looong lists
        if attrlist is not None:
            if attrlist:
                res = [
                    (p.id, {k: get_attribute(p, k) for k in attrlist})
                    for p in query.all()
                ]
            # empty attrlist, so we take all attributes
            else:
                def merged_attrs(p):
                    # merge fixed attributes and dynamic attributes from ``data``
                    attrs = {k: get_attribute(p, k) for k in fixed_attrs if k != 'id'}
                    attrs.update(**{k: get_attribute(p, k) for k in p.data})
                    return attrs

                res = [(p.id, merged_attrs(p)) for p in query.all()]
            res
        else:
            res = [p.id for p in query.all()]

        if exact_match and not res:
            raise ValueError('no entries found')
        return res

    @default
    def create(self, _id, **kw):
        raise NotImplementedError()

    @default
    def __call__(self):
        if use_tm():
            self.session.flush()
        else:
            self.session.commit()


ENCODING = 'utf-8'


class AuthenticationBehavior(Behavior):
    """Handles password authentication for ugm contract:

    - the plumbed class implements the IUsers interface
    - the plumbed class implements get_hashed_pw(id: str) and
      set_hashed_pw(id: str, hpw: str)
    """
    salt_len = default(8)
    hash_func = default(hashlib.sha256)

    def on_authenticated(self, id, **kw):
        """Can be overriden to do after-authentication stuff.
        """

    @override
    def authenticate(self, id=None, pw=None):
        # cannot authenticate user with unset password.
        if not id or not pw:
            return False

        if id not in self:
            return False

        hpw = self.get_hashed_pw(id)
        if hpw:
            authenticated = self._chk_pw(pw, hpw)
            if authenticated:
                self.on_authenticated(id)

            return authenticated
        else:
            return False

    @override
    def passwd(self, id, oldpw, newpw):
        if id not in self:
            raise ValueError(u"User with id '{}' does not exist.".format(id))
        if oldpw is not None:
            if not self._chk_pw(oldpw, self.get_hashed_pw(id)):
                raise ValueError('Old password does not match.')
        hpw = self.hash_passwd(newpw)
        self.set_hashed_pw(id, hpw)
        self()

    @default
    def hash_passwd(self, newpw):
        salt = os.urandom(self.salt_len)
        newpw = newpw.encode(ENCODING) \
            if isinstance(newpw, UNICODE_TYPE) \
            else newpw
        hashed = base64.b64encode(self.hash_func(newpw + salt).digest() + salt)
        return hashed.decode()

    @default
    def _chk_pw(self, plain, hashed):
        hashed = base64.b64decode(hashed)
        salt = hashed[-self.salt_len:]
        plain = plain.encode(ENCODING) \
            if isinstance(plain, UNICODE_TYPE) \
            else plain
        return hashed == self.hash_func(plain + salt).digest() + salt

    @default
    def get_hashed_pw(self, id):
        msg = 'get_hashed_pw(id: str) -> str must be implemented'
        raise NotImplementedError(msg)

    @default
    def set_hashed_pw(self, id, hpw):
        msg = 'set_hashed_pw(id: str, hpw: str) must be implemented'
        raise NotImplementedError(msg)


class UsersBehavior(PrincipalsBehavior, BaseUsers):
    record_class = default(SQLUser)

    @default
    def id_for_login(self, login):
        try:
            # Searchterm has to be enclosed in doublequotes to work on JSON fields
            searchterm = '"%s"' % login
            field_name = SQLUser.login
            if self.session.bind.dialect.name == 'sqlite':
                # If the key to the json field is variable we need a special
                # treatment for sqlite
                field_name = ('$.' + field_name).cast(String)
            res = self.session.query(SQLUser).filter(
                SQLUser.data[field_name].cast(String) == searchterm
            ).one()
            return res.id
        except NoResultFound:
            # if we dont find a login field, fall back assuming id is login
            return login

    @default
    def __getitem__(self, id, default=None):
        try:
            sqluser = self.session.query(SQLUser).filter(SQLUser.id == id).one()
        except NoResultFound:
            raise KeyError(id)
        return User(self, sqluser)

    @default
    def __delitem__(self, id):
        try:
            sqluser = self.session.query(SQLUser).filter(SQLUser.id == id).one()
            self.session.delete(sqluser)
        except NoResultFound:
            raise KeyError(id)

    @default
    def __iter__(self):
        users = self.session.query(SQLUser)
        return iter(map(lambda u: u.id, users))

    @default
    def __setitem__(self, key, value):
        msg = 'users can only be added using the create() method'
        raise NotImplementedError(msg)

    @default
    def create(self, _id, **kw):
        login = kw.pop('login', None)
        for name, value in kw.items():
            if value and name in self.ugm.binary_attrs:
                kw[name] = base64.b64encode(value).decode()
        sqluser = SQLUser(id=_id, login=login, data=kw)
        self.session.add(sqluser)
        self.session.flush()
        return self[_id]

    @default
    def get_hashed_pw(self, id):
        user = self[id]
        return user.record.password

    @default
    def set_hashed_pw(self, id, hpw):
        user = self[id]
        user.record.password = hpw

    @default
    def passwd(self, id, old, new):
        self[id].passwd(old, new)

    @default
    def on_authenticated(self, id, **kw):
        if self.ugm.log_auth:
            user = self[id]
            now = datetime.now()
            if user.record.first_login is None:
                user.record.first_login = now
            user.record.last_login = now

    @default
    def invalidate(self, key=None, *a, **kw):
        self.parent.invalidate(key='users')


@plumbing(
    UsersBehavior,
    AuthenticationBehavior,
    NodeChildValidate,
    Nodespaces,
    Adopt,
    Attributes,
    Nodify,
    SQLSession,
    DefaultInit)
class Users(object):
    pass


class GroupsBehavior(PrincipalsBehavior, BaseGroups):
    record_class = default(SQLGroup)

    @default
    def create(self, _id, **kw):
        for name, value in kw.items():
            if value and name in self.ugm.binary_attrs:
                kw[name] = base64.b64encode(value).decode()
        sqlgroup = SQLGroup(id=_id, data=kw)
        self.session.add(sqlgroup)
        self.session.flush()
        return self[_id]

    @default
    def __getitem__(self, id, default=None):
        try:
            sqlgroup = self.session\
                .query(SQLGroup)\
                .filter(SQLGroup.id == id)\
                .one()
        except NoResultFound:
            raise KeyError(id)
        return Group(self, sqlgroup)

    @default
    def __delitem__(self, id):
        try:
            sqlgroup = self.session\
                .query(SQLGroup)\
                .filter(SQLGroup.id == id)\
                .one()
            self.session.delete(sqlgroup)
        except NoResultFound:
            raise KeyError(id)

    @default
    def __iter__(self):
        groups = self.session.query(SQLGroup)
        return iter(map(lambda u: u.id, groups))

    @default
    def __setitem__(self, key, value):
        msg = 'groups can only be added using the create() method'
        raise NotImplementedError(msg)

    @default
    def invalidate(self, key=None, *a, **kw):
        self.parent.invalidate(key='groups')


@plumbing(
    GroupsBehavior,
    NodeChildValidate,
    Nodespaces,
    Adopt,
    Attributes,
    Nodify,
    SQLSession,
    DefaultInit)
class Groups(object):
    pass


class UgmBehavior(BaseUgm):
    users = default(None)
    groups = default(None)
    user_attrs = default([])
    group_attrs = default([])
    binary_attrs = default([])
    log_auth = default(False)

    @override
    def __init__(self, name, parent, user_attrs,
                 group_attrs, binary_attrs, log_auth):
        self.__name__ = name
        self.__parent__ = parent
        self.users = Users('users', self)
        self.groups = Groups('groups', self)
        self.user_attrs = user_attrs
        self.group_attrs = group_attrs
        self.binary_attrs = binary_attrs
        self.log_auth = log_auth

    @default
    def __call__(self):
        if use_tm():
            self.session.flush()
        else:
            self.session.commit()

    @default
    def add_role(self, role, principal):
        principal.add_role(role)

    @default
    def remove_role(self, role, principal):
        principal.remove_role(role)

    @default
    def roles(self, principal):
        return principal.roles

    @default
    def __getitem__(self, k):
        return getattr(self, k)

    @default
    def __iter__(self, k):
        return iter(['users', 'groups'])

    @default
    def __setitem__(self, k, v):
        raise NotImplementedError('``__setitem__`` not in cone.sql.ugm.Ugm')

    @default
    def __delitem__(self, k, v):
        raise NotImplementedError('``__delitem__`` not in cone.sql.ugm.Ugm')

    @default
    def invalidate(self, key=None):
        if not key:
            self.users = Users('users', self)
            self.groups = Groups('groups', self)
            return
        if key == 'users':
            self.users = Users('users', self)
            return
        if key == 'groups':
            self.groups = Groups('groups', self)
            return
        raise KeyError(key)


@plumbing(
    UgmBehavior,
    Nodify,
    SQLSession)
class Ugm(object):
    pass



from cone.sql import get_session
from cone.sql import testing
from cone.sql.acl import PrincipalRoleRecord
from cone.sql.acl import SQLPrincipalACL
from cone.sql.acl import SQLPrincipalRoles
from node.base import BaseNode
from node.interfaces import IUUID
from node.tests import NodeTestCase
from plumber import plumbing
from pyramid.security import Allow
from zope.interface import implementer
import uuid as uuid_module


@plumbing(SQLPrincipalACL)
class InvalidSQLPrincipalACLNode(BaseNode):
    pass

    @property
    def __acl__(self):
        return []  # pragma nocover


@implementer(IUUID)
@plumbing(SQLPrincipalACL)
class SQLPrincipalACLNode(BaseNode):
    uuid = uuid_module.UUID('1a82fa87-08d6-4e48-8bc2-97ee5a52726d')

    @property
    def __acl__(self):
        return [
            (Allow, 'role:editor', ['edit']),
            (Allow, 'role:manager', ['manage']),
        ]


class TestACL(NodeTestCase):
    layer = testing.sql_layer

    def test_SQLPrincipalACL(self):
        # Create request to ensure sql session available::
        request = self.layer.new_request()
        session = get_session(request)

        # Create invalid node and read ``principal_roles``
        node = InvalidSQLPrincipalACLNode()
        self.assertRaises(RuntimeError, lambda: node.principal_roles)

        # Create node and read ``principal_roles``
        node = SQLPrincipalACLNode()
        principal_roles = node.principal_roles

        self.assertTrue(isinstance(principal_roles, SQLPrincipalRoles))
        self.assertEqual(principal_roles.name, 'principal_roles')

        self.assertEqual([_ for _ in principal_roles], [])
        self.assertEqual(principal_roles._roles_for('someuser'), [])

        # check number of entries in table
        self.assertEqual(session.query(PrincipalRoleRecord).count(), 0)

        # check __acl__ on node
        self.assertEqual(node.__acl__, [
            ('Allow', 'role:editor', ['edit']),
            ('Allow', 'role:manager', ['manage'])
        ])

        # Set some local roles for node
        principal_roles['someuser'] = ['manager']
        principal_roles['otheruser'] = ['editor']
        principal_roles['group:some_group'] = ['editor', 'manager']

        # check number of entries in table
        self.assertEqual(session.query(PrincipalRoleRecord).count(), 4)

        # check principal roles via API
        self.assertEqual(
            sorted([_ for _ in principal_roles]),
            ['group:some_group', 'otheruser', 'someuser']
        )
        self.assertEqual(principal_roles['someuser'], ['manager'])
        self.assertEqual(
            sorted(principal_roles['group:some_group']),
            ['editor', 'manager']
        )

        # check __acl__ on node
        acl = node.__acl__
        self.assertEqual(len(acl), 5)
        self.assertEqual(acl[0][0], 'Allow')
        self.assertEqual(acl[0][1], 'someuser')
        self.assertEqual(acl[0][2], ['manage'])

        self.assertEqual(acl[1][0], 'Allow')
        self.assertEqual(acl[1][1], 'otheruser')
        self.assertEqual(acl[1][2], ['edit'])

        self.assertEqual(acl[2][0], 'Allow')
        self.assertEqual(acl[2][1], 'group:some_group')
        self.assertEqual(sorted(acl[2][2]), ['edit', 'manage'])

        self.assertEqual(acl[3][0], 'Allow')
        self.assertEqual(acl[3][1], 'role:editor')
        self.assertEqual(acl[3][2], ['edit'])

        self.assertEqual(acl[4][0], 'Allow')
        self.assertEqual(acl[4][1], 'role:manager')
        self.assertEqual(acl[4][2], ['manage'])

        # Modify some local roles
        principal_roles['group:some_group'] = ['viewer', 'admin', 'manager']

        # check number of entries in table
        self.assertEqual(session.query(PrincipalRoleRecord).count(), 5)

        # check principal roles via API
        self.assertEqual(
            sorted(principal_roles['group:some_group']),
            ['admin', 'manager', 'viewer']
        )
        self.assertEqual(
            sorted(principal_roles.keys()),
            ['group:some_group', 'otheruser', 'someuser']
        )
        self.assertEqual(
            sorted(principal_roles['group:some_group']),
            ['admin', 'manager', 'viewer']
        )
        self.assertEqual(principal_roles['otheruser'], ['editor'])
        self.assertEqual(principal_roles['someuser'], ['manager'])

        # Create another node and set some local roles for it
        node = SQLPrincipalACLNode()
        node.uuid = uuid_module.UUID('67140d49-6a4d-4d72-9171-ad764131c880')
        principal_roles = node.principal_roles
        principal_roles['otheruser'] = ['editor', 'admin']
        principal_roles['group:some_group'] = ['editor', 'manager']

        # check number of entries in table
        self.assertEqual(session.query(PrincipalRoleRecord).count(), 9)

        # check principal roles via API
        self.assertEqual(
            sorted(principal_roles.keys()),
            ['group:some_group', 'otheruser']
        )
        self.assertEqual(
            sorted(principal_roles['group:some_group']),
            ['editor', 'manager']
        )
        self.assertEqual(
            sorted(principal_roles['otheruser']),
            ['admin', 'editor']
        )

        # Delete local roles
        del principal_roles['group:some_group']
        self.assertEqual(principal_roles.keys(), ['otheruser'])

        # check number of entries in table
        self.assertEqual(session.query(PrincipalRoleRecord).count(), 7)

        # check principal roles via API
        self.assertEqual(
            sorted(principal_roles['otheruser']),
            ['admin', 'editor']
        )

        # Check if local roles of first node still sane
        node = SQLPrincipalACLNode()
        principal_roles = node.principal_roles
        self.assertEqual(
            sorted(principal_roles.keys()),
            ['group:some_group', 'otheruser', 'someuser']
        )
        self.assertEqual(
            sorted(principal_roles['group:some_group']),
            ['admin', 'manager', 'viewer']
        )
        self.assertEqual(principal_roles['otheruser'], ['editor'])
        self.assertEqual(principal_roles['someuser'], ['manager'])



from setuptools import find_packages
from setuptools import setup
from setuptools.command.test import test
import os


def read_file(name):
    with open(os.path.join(os.path.dirname(__file__), name)) as f:
        return f.read()


version = '0.3'
shortdesc = 'SQLAlchemy integration for cone.app'
longdesc = '\n\n'.join([read_file(name) for name in [
    'README.rst',
    'CHANGES.rst',
    'LICENSE.rst'
]])


class Test(test):

    def run_tests(self):
        from cone.sql import tests
        tests.run_tests()


setup(
    name='cone.sql',
    version=version,
    description=shortdesc,
    long_description=longdesc,
    classifiers=[
        'Environment :: Web Environment',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    keywords='node pyramid cone web',
    author='Robert Niederreiter',
    author_email='dev@bluedynamics.com',
    url=u'https://github.com/bluedynamics/cone.sql',
    license='Simplified BSD',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['cone'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'zope.sqlalchemy',
        'node.ext.ugm',
        'pyramid_tm',
        'pyramid_retry',
        'cone.app',
    ],
    extras_require=dict(
        ugm=[
            'cone.ugm'
        ],
        test=[
            'zope.testrunner',
            'cone.ugm'
        ]
    ),
    tests_require=[
        'zope.testrunner',
        'cone.ugm'
    ],
    cmdclass=dict(test=Test),
    entry_points="""\
    [paste.filter_app_factory]
    session = cone.sql:make_app
    """
)



import sys
import unittest


def test_suite():
    from cone.sql.tests import test_sql
    from cone.sql.tests import test_model
    from cone.sql.tests import test_acl
    from cone.sql.tests import test_ugm

    suite = unittest.TestSuite()

    suite.addTest(unittest.findTestCases(test_sql))
    suite.addTest(unittest.findTestCases(test_model))
    suite.addTest(unittest.findTestCases(test_acl))
    suite.addTest(unittest.findTestCases(test_ugm))

    try:
        from cone.sql.tests import test_cone_ugm
        suite.addTest(unittest.findTestCases(test_cone_ugm))
    except ImportError:
        pass

    return suite


def run_tests():
    from zope.testrunner.runner import Runner

    runner = Runner(found_suites=[test_suite()])
    runner.run()
    sys.exit(int(runner.failed))


if __name__ == '__main__':
    run_tests()



from cone.app.model import AppNode
from cone.sql import get_session
from cone.sql import use_tm
from node.behaviors import Adopt
from node.behaviors import Attributes
from node.behaviors import DefaultInit
from node.behaviors import Lifecycle
from node.behaviors import NodeAttributes
from node.behaviors import Nodespaces
from node.behaviors import Nodify
from node.interfaces import ICallable
from node.interfaces import IStorage
from plumber import Behavior
from plumber import default
from plumber import finalize
from plumber import override
from plumber import plumbing
from pyramid.threadlocal import get_current_request
from sqlalchemy import inspect
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.types import CHAR
from sqlalchemy.types import TypeDecorator
from zope.interface import implementer
import sys
import uuid


###############################################################################
# Compat
###############################################################################

IS_PY2 = sys.version_info[0] < 3
UNICODE_TYPE = unicode if IS_PY2 else str


###############################################################################
# SQLAlchemy data types
###############################################################################

class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses Postgresql's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    http://docs.sqlalchemy.org/en/rel_0_8/core/types.html#backend-agnostic-guid-type
    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % int(uuid.UUID(value))
            else:
                # hexstring
                return "%.32x" % int(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return uuid.UUID(value)


###############################################################################
# SQL table storage
###############################################################################

@implementer(IStorage, ICallable)
class SQLTableStorage(Behavior):
    # SQL alchemy model class
    record_class = default(None)
    # SQL alchemy session
    session = default(None)
    # factory for node children
    child_factory = default(None)
    # map SQL alchemy data types to callables converting values to expected
    # type
    data_type_converters = default({
        GUID: uuid.UUID,
        String: UNICODE_TYPE,
        Integer: int,
    })

    @default
    @property
    def primary_key(self):
        if not hasattr(self, '_primary_key'):
            self._primary_key = inspect(self.record_class).primary_key
        return self._primary_key

    @default
    def _convert_primary_key(self, name):
        # XXX: multiple primary key support
        primary_key = self.primary_key[0]
        primary_key_type = primary_key.type.__class__
        try:
            converter = self.data_type_converters[primary_key_type]
            primary_key_value = converter(name)
            return primary_key_value
        except Exception as e:
            msg = (
                'Failed to convert node name to expected primary key '
                'data type: {}'
            ).format(e)
            raise KeyError(msg)

    @finalize
    def __setitem__(self, name, value):
        # XXX: multiple primary key support
        primary_key = self.primary_key[0]
        primary_key_value = self._convert_primary_key(name)
        attrs = value.attrs
        if not attrs[primary_key.name]:
            attrs[primary_key.name] = primary_key_value
        if primary_key_value != attrs[primary_key.name]:
            msg = (
                'Node name must match primary key attribute value: {} != {}'
            ).format(primary_key_value, attrs[primary_key.name])
            raise KeyError(msg)
        session = self.session
        query = session.query(self.record_class).filter(
            getattr(self.record_class, primary_key.name) == primary_key_value)
        if not session.query(query.exists()).scalar():
            session.add(value.record)
        else:
            record = query.first()
            for k, v in value.attrs.items():
                setattr(record, k, v)
            value.record = value.attrs.record = record

    @finalize
    def __getitem__(self, name):
        # XXX: multiple primary key support
        primary_key = self.primary_key[0]
        primary_key_value = self._convert_primary_key(name)
        session = self.session
        query = session.query(self.record_class)
        record = query.filter(
            getattr(self.record_class, primary_key.name) == primary_key_value
        ).first()
        if record is None:
            # traversal expects ``KeyError`` before looking up views.
            raise KeyError(name)
        return self.child_factory(name, self, record)

    @finalize
    def __delitem__(self, name):
        child = self[name]
        session = self.session
        session.delete(child.record)

    @finalize
    def __iter__(self):
        # XXX: multiple primary key support
        primary_key = self.primary_key[0]
        session = self.session
        result = session.query(getattr(self.record_class, primary_key.name))
        for recid in result.all():
            yield str(recid[0])

    @finalize
    def __call__(self):
        if use_tm():
            self.session.flush()
        else:
            self.session.commit()


###############################################################################
# SQL row storage
###############################################################################

class SQLRowNodeAttributes(NodeAttributes):

    def __init__(self, name, parent, record):
        NodeAttributes.__init__(self, name, parent)
        self.record = record

    @property
    def _columns(self):
        return inspect(self.record.__class__).attrs.keys()

    def __setitem__(self, name, value):
        if name in self:
            setattr(self.record, name, value)
        else:
            raise KeyError('Unknown attribute: {}'.format(name))

    def __getitem__(self, name):
        if name in self:
            return getattr(self.record, name)
        raise KeyError('Unknown attribute: {}'.format(name))

    def __delitem__(self, name):
        raise KeyError('Deleting of attributes not allowed')

    def __iter__(self):
        return iter(self._columns)

    def __contains__(self, name):
        return name in self._columns


@implementer(IStorage, ICallable)
class SQLRowStorage(Behavior):
    # SQL alchemy model class
    record_class = default(None)
    # SQL alchemy session
    session = default(None)

    @override
    def __init__(self, name=None, parent=None, record=None):
        self.__name__ = name
        self.__parent__ = parent
        self._new = False
        if record is None:
            self._new = True
            record = self.record_class()
        self.record = record

    @override
    def attributes_factory(self, name, parent):
        return SQLRowNodeAttributes(name, parent, self.record)

    @finalize
    def __setitem__(self, name, value):
        raise KeyError(name)

    @finalize
    def __getitem__(self, name):
        raise KeyError(name)

    @finalize
    def __delitem__(self, name):  # pragma: no cover
        # not reached by default SQLRowNode, KeyError already thrown in
        # Lifecycle plumbing Behavior
        raise KeyError(name)

    @finalize
    def __iter__(self):
        return iter([])

    @finalize
    def __call__(self):
        session = self.session
        if self._new:
            session.add(self.record)
            self._new = False
        if use_tm():
            self.session.flush()
        else:
            self.session.commit()


###############################################################################
# SQL session provider
###############################################################################

class SQLSession(Behavior):
    """Behavior providing SQLAlchemy session from pyramid request.
    """

    @finalize
    @property
    def session(self):
        return get_session(get_current_request())


###############################################################################
# Application node basics
###############################################################################

@plumbing(
    AppNode,
    Adopt,
    DefaultInit,
    Nodify,
    Lifecycle,
    SQLSession,
    SQLTableStorage)
class SQLTableNode(object):
    """Basic SQL table providing node.
    """


@plumbing(
    AppNode,
    Nodespaces,
    Attributes,
    Nodify,
    Lifecycle,
    SQLSession,
    SQLRowStorage)
class SQLRowNode(object):
    """Basic SQL row providing node.
    """
